/*
 * Copyright (c) [2004-2009] Novell, Inc.
 *
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, contact Novell, Inc.
 *
 * To contact Novell about this file by physical or electronic mail, you may
 * find current contact information at www.novell.com.
 */

/*
  Textdomain    "storage"
*/

#include <iostream>
#include <sstream>
#include <locale>
#include <boost/algorithm/string.hpp>

#include "y2storage/DmmultipathCo.h"
#include "y2storage/Dmmultipath.h"
#include "y2storage/SystemCmd.h"
#include "y2storage/AppUtil.h"
#include "y2storage/Storage.h"
#include "y2storage/StorageDefines.h"

using namespace std;
using namespace storage;


    CmdMultipath::CmdMultipath()
    {
	SystemCmd c(MULTIPATHBIN " -d -v 2+ -ll");
	if (c.retcode() != 0 || c.numLines() == 0)
	    return;

	vector<string> lines;
	c.getStdout(lines);
	vector<string>::const_iterator it1 = lines.begin();

	while (it1 != lines.end())
	{
	    Entry entry;

	    y2mil("mp line:" << *it1);

	    string name = extractNthWord(0, *it1);
	    y2mil("mp name:" << name);

	    list<string> tmp = splitString(extractNthWord(2, *it1, true), ",");
	    if (tmp.size() >= 2)
	    {
		list<string>::const_iterator it2 = tmp.begin();
		entry.vendor = boost::trim_copy(*it2++, locale::classic());
		entry.model = boost::trim_copy(*it2++, locale::classic());
	    }

	    ++it1;

	    if (it1 != lines.end())
		++it1;

	    while (it1 != lines.end())
	    {
		if (it1->empty() || isalnum((*it1)[0]))
		    break;

		if (boost::starts_with(*it1, "| `-") || boost::starts_with(*it1, "  `-") ||
		    boost::starts_with(*it1, "  |-"))
		{
		    string tmp = it1->substr(5);
		    y2mil("mp element:" << tmp);
		    string dev = "/dev/" + extractNthWord(1, tmp);
		    if (find(entry.devices.begin(), entry.devices.end(), dev) == entry.devices.end())
			entry.devices.push_back(dev);
		}

		++it1;
	    }

	    data[name] = entry;
	}

	for (const_iterator it = data.begin(); it != data.end(); ++it)
	    y2mil("data[" << it->first << "] -> vendor:" << it->second.vendor <<
		  " model:" << it->second.model << " devices:" << it->second.devices);
    }


    list<string>
    CmdMultipath::getEntries() const
    {
	list<string> ret;
	for (const_iterator it = data.begin(); it != data.end(); ++it)
	    ret.push_back(it->first);
	return ret;
    }


    bool
    CmdMultipath::getEntry(const string& name, Entry& entry) const
    {
	const_iterator it = data.find(name);
	if (it == data.end())
	    return false;

	entry = it->second;
	return true;
    }


DmmultipathCo::DmmultipathCo( Storage * const s, const string& Name, ProcPart& ppart ) :
    DmPartCo(s, "/dev/mapper/"+Name, staticType(), ppart )
{
    DmPartCo::init( ppart );
    getMultipathData( Name );
    y2deb("constructing dmmultipath co " << Name);
}


DmmultipathCo::~DmmultipathCo()
{
    y2deb("destructed multipath co " << dev);
}


void
DmmultipathCo::getMultipathData(const string& name)
{
    y2mil("name:" << name);

    CmdMultipath::Entry entry;
    if (CmdMultipath().getEntry(name, entry))
    {
	vendor = entry.vendor;
	model = entry.model;

	Pv *pv = new Pv;
	for (list<string>::const_iterator it = entry.devices.begin(); it != entry.devices.end(); ++it)
	{
	    pv->device = *it;
	    addPv(pv);
	}
	delete pv;
    }
}


void
DmmultipathCo::setUdevData(const list<string>& id)
{
    y2mil("disk:" << nm << " id:" << id);
    udev_id = id;
    udev_id.erase(remove_if(udev_id.begin(), udev_id.end(), find_begin("dm-")), udev_id.end());
    udev_id.sort();
    y2mil("id:" << udev_id);

    DmPartCo::setUdevData(udev_id);

    DmmultipathPair pp = dmmultipathPair();
    for( DmmultipathIter p=pp.begin(); p!=pp.end(); ++p )
	{
	p->addUdevData();
	}
}


void
DmmultipathCo::newP( DmPart*& dm, unsigned num, Partition* p )
    {
    y2mil( "num:" << num );
    dm = new Dmmultipath( *this, num, p );
    }


void
DmmultipathCo::addPv(Pv*& p)
{
    PeContainer::addPv(p);
    if (!deleted())
	getStorage()->setUsedBy(p->device, UB_DMMULTIPATH, name());
    p = new Pv;
}


void
DmmultipathCo::activate(bool val)
{
    y2mil("old active:" << active << " val:" << val);

    if (active != val)
    {
	bool lvm_active = LvmVg::isActive();
	LvmVg::activate(false);

	SystemCmd c;
	if (val)
	{
	    Dm::activate(true);

	    c.execute(MODPROBEBIN " dm-multipath");
	    c.execute(MULTIPATHBIN);
	    sleep(1);
	    c.execute(MULTIPATHDBIN);
	}
	else
	{
	    c.execute(MULTIPATHDBIN " -F");
	    sleep(1);
	    c.execute(MULTIPATHBIN " -F");
	}

	LvmVg::activate(lvm_active);

	active = val;
    }
}


    bool
    DmmultipathCo::isActivated(const string& name)
    {
	SystemCmd c(DMSETUPBIN " table " + quote(name));
	return c.retcode() == 0 && c.numLines() >= 1 && isdigit(c.stdout()[0]);
    }


    list<string>
    DmmultipathCo::getMultipaths()
    {
	list<string> l;

	list<string> entries = CmdMultipath().getEntries();
	for (list<string>::const_iterator it = entries.begin(); it != entries.end(); ++it)
	{
	    if (isActivated(*it))
		l.push_back(*it);
	    else
		y2mil("ignoring inactive dmmultipath " << *it);
	}

	if (!l.empty())
	    active = true;

	y2mil("detected multipaths " << l);
	return l;
    }


string DmmultipathCo::setDiskLabelText( bool doing ) const
    {
    string txt;
    string d = nm;
    if( doing )
        {
        // displayed text during action, %1$s is replaced by multipath name (e.g. 3600508b400105f590000900000300000),
	// %2$s is replaced by label name (e.g. msdos)
        txt = sformat( _("Setting disk label of multipath disk %1$s to %2$s"),
		       d.c_str(), labelName().c_str() );
        }
    else
        {
        // displayed text before action, %1$s is replaced by multipath name (e.g. 3600508b400105f590000900000300000),
	// %2$s is replaced by label name (e.g. msdos)
        txt = sformat( _("Set disk label of multipath disk %1$s to %2$s"),
		      d.c_str(), labelName().c_str() );
        }
    return( txt );
    }


void DmmultipathCo::getInfo( DmmultipathCoInfo& tinfo ) const
    {
    DmPartCo::getInfo( info );
    tinfo.p = info;
    tinfo.vendor = vendor;
    tinfo.model = model;
    }

namespace storage
{

std::ostream& operator<<(std::ostream& s, const DmmultipathCo& d)
{
    s << *((DmPartCo*)&d);
    s << " Vendor:" << d.vendor
      << " Model:" << d.model;
    return s;
}

}

string DmmultipathCo::getDiffString( const Container& d ) const
{
    string log = DmPartCo::getDiffString(d);
    const DmmultipathCo * p = dynamic_cast<const DmmultipathCo*>(&d);
    if (p)
    {
	if (vendor != p->vendor)
	    log += " vendor:" + vendor + "-->" + p->vendor;
	if (model != p->model)
	    log += " model:" + model + "-->" + p->model;
    }
    return log;
}

bool DmmultipathCo::equalContent( const Container& rhs ) const
{
    bool ret = Container::equalContent(rhs);
    if (ret)
    {
	const DmmultipathCo *p = dynamic_cast<const DmmultipathCo*>(&rhs);
	ret = p && DmPartCo::equalContent(*p) &&
	    vendor == p->vendor && model == p->model;
    }
    return ret;
}

DmmultipathCo::DmmultipathCo( const DmmultipathCo& rhs ) : DmPartCo(rhs)
{
    vendor = rhs.vendor;
    model = rhs.model;
}

void DmmultipathCo::logData( const string& Dir ) {}

bool DmmultipathCo::active = false;
